/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

 
//                          ptt version: 1.0


//  xxxxxxxxxxxxxxx    global variable declarations  xxxxxxxxxxxxxxxx

  
// These variables below can be directly modified by the user on the options form: 

var tpath = 1;   //  path type: Sine (1), Rectangle (2), line(3), Arc(4)       
 
var direc = 1;   // direction: Left-Right(1), Right-Left(2), Up-Down(3), Down-Up(4). Range 1..4  

var radius = 2;	 // path radius in percents of canvas width/height. units: percents converted into pixels (px) in setPar()

var period = 1;    // period => number of sine or rectangle cycles within canvas region bounds

var timer_enabled = false;   // if true => timer is enabled. if false timer is disabled. Exactly: the timer is always on

var timer = 60;     // Task time in seconds if clock = 1000ms (default).
 
var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 

// Other global constants and variables:

var timer_id = null;   // timer_id is the identifier of the repeated action you want to cancel. This ID is returned from setInterval().

var is_task = false;    // task in progress or not. is_task is true for all events between startTask() and theEnd() events

var dirty = false;  // if  (dirty==true) => user changed any options in form 

var is_storage = false; // local storage is enabled on user browser or not

var cnt = 0;    	  // global counter for Timer event 

var point = 0;		// number of points during task (0 or 1)

var err = 0;      // error number in task

var xMouse, yMouse;   // mouse pointer (pointing device) coordinates in px 

var startX, startY, endX, endY;    // start red circle coord (x,y), end blue circle coord (x,y)
		
var startTime;    // the time the task began. init in startTask()

var is_path= false;  // white path is visible on screen (true) or not (false)

var bounds;   // rectangle object => canvas bounds for path. bounds = {left : x1, top : x2, right : x3, bottom : x4};


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
		xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */


// application init [check storage, show start panel..] 
function init() 
{     
  if (timer_id)  {   
    window.clearInterval(timer_id);  
  }  
  timer_id= setInterval(function () {appTimer()},100);
	is_storage= isStorage() && storageAvailable('localStorage');
  if (is_storage)  {
    document.getElementById('store_group').style.display= 'block';
		document.getElementById('no_storage').style.display= 'none';
  }				  
	setIns();		
	showStart();
}


// application timer handler function
function appTimer()  
{    
	if (timer_enabled) 			// if (timer_enabled /*&& is_task)
	{ 
		cnt++;  
		if (cnt >= 10 * timer)  {     // task time is limited
			theEnd(false);
		}	
	}  
	// else: nop, idle, do nothing 
}	


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
     return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// raised when page is resized. see html: <body onresize="pageResize(true || false)"
// stretch canvas 
function pageResize(par)  
{ 
	if (is_task || par)  { 
		let c= document.getElementById('can');
		c.width= window.innerWidth;  
		c.style.width= c.width + 'px';	     
		c.height= window.innerHeight;   
		c.style.height= c.height + 'px';  // https://css-tricks.com/whats-the-difference-between-width-height-in-css-and-width-height-html-attributes/ 							
		if (!par) { 
			theEnd(true);	  // resize error		
    } 		 
	}	
}	


// show start panel 
function showStart()  
{      
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('center-xy');  // ex 'child' class    
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  document.getElementById('hide_option').style= 'text-align:center';  // bug? auto-remove style when element.style.display="none"
  document.getElementById('instructions').style.display= 'none';     
  document.getElementById('store_oper').style.display= 'none';
	if (is_storage && localStorage.getItem('ptt-v1.0')) 
  {
    let cnt= JSON.parse(localStorage.getItem('ptt-v1.0')).length;                                                        
    document.getElementById('store_cnt').textContent=  'You have a total of ' + cnt  + ' results stored in your web browser.';
    if (cnt > 0) {
			document.getElementById('store_oper').style.display= 'table';  // Watch this line: this was a nasty bug 
		}
	}
  dirty= false;
  //document.getElementById('is_valid').textContent='';  // tmp reset validator message
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  
	document.getElementById('options').style.display = 'none'; 
  document.getElementById('start').classList.add('center-xy');   // ex 'child' class 
  document.getElementById('show_option').style.display = 'block';   
  document.getElementById('start_div').style.display = 'block';          
  document.getElementById('hide_option').style.display = 'none';   
  //document.getElementById('start_div').style= 'text-align:center';  // bug?  (firefox auto remove class "center" when set style.display.none;)
  document.getElementById('show_option').style= 'text-align:center';  
  document.getElementById('instructions').style.display= 'block';
  
	refMsg(); 
  
	if ( tpath == 2 ) {
    checkDel();
	}		
	setIns();	
}


// set instructions text depending on user selected task options
function setIns()  
{
  document.getElementById('instructions').style.display= 'block';  
	if ( document.getElementById('timer_enabled').checked )  {
		document.getElementById('ins_timer').style.display= 'block';
		document.getElementById('ins_timer').textContent= 'The duration of the task is about ' + timer.toString() + ' seconds or ' + stringTime(timer) + '.'; 		
	}
	else  {
		document.getElementById('ins_timer').style.display= 'none';
	}
}	


// refresh Options (update *_msg label) => give user short option HELP
function refOpt(e)  
{   
  dirty=true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';               
	switch (my_id)
	{
		case 'tpath':		  
		  document.getElementById('p_period').style.display = (e.value == '1' || e.value == '2') ? 'block' : 'none';
			document.getElementById('path_warning').style.display = (e.value == '1' || e.value == '2') ? 'block' : 'none';
			break;

    case 'direc':                                                                                                                                            
      if (e.value == '1')  {
        msg.textContent= 'The start is on the left side of the screen and the finish is on the right side of the screen.';
      }
      else if (e.value == '2')  {
        msg.textContent= 'The start is on the right side of the screen and the finish is on the left side of the screen.';
      }
			else if (e.value == '3')  {
        msg.textContent= 'The start is at the top of the screen and the finish is at the bottom of the screen.';
      }
      else  {
        msg.textContent= 'The start is at the bottom of the screen and the finish is at the top of the screen.'; 
      }
      break;    
			        
    case 'radius':                                                                                                                                     
      msg.textContent= e.options[e.selectedIndex].value + '%';
      break;    							
	
		case 'period':
			if (e.validity.valid)  {
				msg.textContent= 'Number of repetitions of the same curve on the screen. Available only for sine and rectangle path. Period of the curve is the length of one cycle of the curve.'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break;
		
		case 'timer_enabled':   
      if (e.checked)  {
        msg.textContent= 'Yes';
        document.getElementById('p_timer').style.display= 'block';
      }
      else  {  
        msg.textContent= 'Timer is disabled';
        document.getElementById('p_timer').style.display= 'none'; 
      }       
      break;
		
    case 'timer':
			if (e.validity.valid)  {
				msg.textContent= stringTime(e.value); 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break;
	
			
		case 'store':                                                                                                                                            
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
 
	}   // switch
	
  refMsg();   
}


// return integer into string hh:mm:ss time format
// input par: total == integer (time in seconds)
function stringTime(total) 
{
  let hh = Math.floor(total/(60*60));         
  let div_min = total % (60 * 60);
  let mm = Math.floor(div_min / 60);       
  let div_sec = div_min % 60;
  let ss = Math.ceil(div_sec);    
  // These lines ensure you always have two-digits. This formats your string to HH:MM:SS
  if (hh < 10)  {hh = "0"+hh;}     
  if (mm < 10) {mm = "0"+mm;}
  if (ss < 10) {ss = "0"+ss;} 
  return hh + ":" + mm + ":" + ss;
} 


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  dirty === true ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// check for validity.valid object in all input type="number"
function optionsValid()  
{ 
  let res= document.getElementById('period').validity.valid && document.getElementById('timer').validity.valid; 	
  return res;              
}


// start task (research) 
function startTask(e)   
{       			
	initLayout();
  
	pageResize(true);	  	// must be called before setPar() => set window.radius
	
	setPar();
	
	is_task= true;     //  see pageResize()
	
  timer_enabled= (document.getElementById('timer_enabled').checked) ? true: false;  // timer is enabled or disabled
  
	drawPath();
	
	drawCircle();
	
	startTime= new Date();  
}


// init visible and invisible html dom elements before start task
function initLayout()
{
	document.getElementById('result_wrapper').style.display= 'none';
  document.getElementById('end').style.display= 'none'; 
  document.getElementById('start').style.display= 'none';               	  
	document.getElementById('can').style.display= 'block';			     
}

 
// set internal application research parameters (global vars)
function setPar()  
{  		 		
	cnt= 0; 
	point= 0;	 
	err= 0;  
	
	if ( document.getElementById('radius').selectedIndex > -1 ) {  // update window.radius. convert percents into pixels
		let r = parseInt ( document.getElementById('radius').value, 10 );
		radius = Math.floor ( r * Math.min( document.getElementById('can').width, document.getElementById('can').height ) / 100 );  // must be integer 
	}	
	if (tpath == 4) {
		period= 0.5;  // period => number of sine cycles within canvas region bounds
	}
	 
	let c = document.getElementById('can'), w = c.width, h = c.height;
	bounds = {left : radius, top : radius, right : w-radius, bottom : h-radius};  

	xMouse= null; 
	yMouse= null;
	
	is_path= false;
}


// init and draw path
function drawPath()
{
  switch (tpath)  {
		case 1:
		case 4:  
			drawSin();
		  break;
		
		case 2:
		  drawRect();
			break;
			
    case 3:
			drawLine();
			break;
	
		}
}


// draw horizontal or vertical path of the sine (or arc) function
// handle path for tpath = sine and tpath == arc
function drawSin()
{
	clear();
		
	let frequency, baseline, amplitude;  
	let hor = (direc == 1 || direc == 2) ? true: false; 
	if ( hor == true )  // horizontal sine function
	{		
		frequency = Math.abs( bounds.right - bounds.left )  / ( 2 * period * Math.PI ); 	// period => number of sine cycles within canvas region bounds
		baseline = bounds.top + ( bounds.bottom - bounds.top ) / 2; 																	
		amplitude = Math.floor( ( bounds.bottom - bounds.top) / 2 - radius ); 
	}	
	else  // vertical sinus function
	{		
		frequency = Math.abs( bounds.bottom - bounds.top ) / ( 2 * period * Math.PI );
		baseline =  bounds.left + ( bounds.right - bounds.left ) / 2; 	 			      																
		amplitude = Math.floor( ( bounds.right - bounds.left ) / 2 - radius );   		
	}	

	let c = document.getElementById('can'), ctx= c.getContext('2d');  			
	ctx.fillStyle= '#fff';    
	let cnt= 0;   // != window.cnt
	if ( hor == true )
	{	
		for (let x= bounds.left; x <= bounds.right; x++)  
		{		 
			let y = baseline + amplitude * Math.sin(cnt/frequency);			
			ctx.beginPath(); 			
	    ctx.arc(x, y, radius, 0, 2 * Math.PI);
			ctx.fill();
			if ( x == bounds.left + radius ) {
				startX= x;   
				startY= y; 
			}
			else if ( x == bounds.right - radius ) {
				endX= x;   
				endY= y; 
			}
			cnt++;			
		}
	}
	else 
	{	
		for (let y= bounds.top; y <= bounds.bottom; y++)  {
			let x = baseline + amplitude * Math.sin(cnt/frequency);				
			ctx.beginPath(); 			
	    ctx.arc(x, y, radius, 0, 2 * Math.PI);
			ctx.fill();
		 	if ( y  == bounds.top + radius ) {
				startX= x;   
				startY= y; 
			}
		  else if ( y == bounds.bottom - radius ) {
				endX= x;   
				endY= y; 
			}
			cnt++;
		}		
	}
 	
	if ( direc == 2 || direc == 4 )  { 
		swap();		
	}
	ctx.fillStyle= '#00f';	
	ctx.beginPath(); 
	ctx.arc(endX, endY, radius, 0, 2*Math.PI);
	ctx.fill(); 	
}	


// draw horizontal or vertical line path. handle path for tpath = 2
function drawRect()
{
	clear();
	
	let c = document.getElementById('can'), ctx= c.getContext('2d');  			
	
	ctx.fillStyle= '#fff';	
	if ( direc == 1 || direc == 2 )  // direction: 1 == left-right or 2 == right-left
	{				
    startX = bounds.left + radius;
		startY = bounds.top + radius; 
		endX = bounds.right - radius;
		endY = startY;
		
		//( 4 + radius ) + 2 * dx  = ( bounds.right - bounds.left) / period  // equation	
		let dx = ( ( bounds.right - bounds.left) / period  - 4 * radius ) / 2; 
		
		let x= bounds.left;
		for (let j= 0; j < period; j++) 
		{    
			ctx.fillRect(x, bounds.top, 2*radius, bounds.bottom - bounds.top );					
			x+= 2 * radius;  // always increase as width parameter in last fillRect() call = 3. par is width
					        
			ctx.fillRect(x, bounds.bottom - 2*radius, dx, 2*radius );
			x+= dx;	 
			  
			ctx.fillRect(x, bounds.top, 2*radius, bounds.bottom - bounds.top );
			x+= 2 * radius;

			ctx.fillRect(x, bounds.top, dx, 2*radius);
   		x+= dx;
		
		}
				
	}	
	else  // direction: 3 => Up-Down or 4 => Down-Up
	{		
		startX = bounds.right - radius;
		startY = bounds.top + radius; 
		endX = startX;
		endY = bounds.bottom - radius;
		
		//( 4 + radius) + 2 * dy  = ( bounds.bottom - bounds.top) / period  // equation
		let dy = ( ( bounds.bottom - bounds.top) / period  - 4 * radius ) / 2; 
    let y = bounds.top;	
		for (let j= 1; j <= period; j++ ) 
		{  
			ctx.fillRect(bounds.left, y, bounds.right-bounds.left, 2*radius);
		  y+= 2 * radius;  // always increase y equal as previous stripe height in fillRect() call => last parameter  		
			
			ctx.fillRect(bounds.left, y, 2*radius, dy);
		  y+= dy;
			 
			ctx.fillRect(bounds.left, y, bounds.right - bounds.left, 2*radius);
			y+= 2 * radius;
	
			ctx.fillRect(bounds.right - 2 * radius, y,  2*radius, dy);
			y+= dy;			
		}
		
	}
		
  if ( direc == 2 || direc == 4 )  { 
		swap();		
	}
	ctx.fillStyle= '#00f';	
	ctx.beginPath(); 
	ctx.arc(endX, endY, radius, 0, 2*Math.PI);
	ctx.fill(); 
 
}	


// draw horizontal or vertical line path. handle path for tpath = 3
function drawLine()
{
	clear();
		
	let c = document.getElementById('can'), ctx= c.getContext('2d');  			
	ctx.fillStyle= '#fff';
	
	if ( direc == 1 || direc == 2 )  // direction: 1 == left-right or 2 == right-left
	{				
		ctx.fillRect( bounds.left, Math.round ( ( bounds.bottom - bounds.top) / 2  - radius), bounds.right - bounds.left, 2*radius );		
		if ( direc == 1 ) {
			startX = bounds.left + radius;
			endX = bounds.right - radius;
		}
		else  {
			startX = bounds.right - radius;
			endX = bounds.left + radius;		
		}
		startY = Math.round((bounds.bottom - bounds.top)/2);
    endY =  startY;				   
	}	
	else  // direction: 3 =>  Up-Down or 4 => Down-Up
	{		
		ctx.fillRect( Math.round ( (bounds.right - bounds.left) / 2  - radius), bounds.top, 2 * radius, bounds.bottom - bounds.top );		
		startX = Math.round ( (bounds.right - bounds.left) / 2);
		endX = startX;
		if ( direc == 3 ) {
			startY= bounds.top + radius;
			endY = bounds.bottom - radius;
    }
    else  {		
			startY = bounds.bottom - radius;
		  endY= bounds.top + radius;
		}
	}	
	
	ctx.fillStyle= '#00f';	
	ctx.beginPath(); 
	ctx.arc(endX, endY, radius, 0, 2*Math.PI);
	ctx.fill(); 
		
}	


// draw start (red) circle 
function drawCircle()
{	
  clear();
	if ( xMouse == null || yMouse == null  )  {
		gotoStart();	
	}		 

	let ctx= document.getElementById('can').getContext('2d');		
	ctx.fillStyle= '#f00';	
	ctx.beginPath(); 
	ctx.arc(xMouse, yMouse, radius, 0, 2*Math.PI);
	ctx.fill(); 	
}


// reset user to start point
function gotoStart()
{
	xMouse= startX;
  yMouse= startY;				
}	


// fill canvas object (screen) with background color
function clear()
{
	let c = document.getElementById('can'), ctx= c.getContext('2d'), w = c.width, h = c.height;	
	ctx.fillStyle= '#e5e5e5';  // Note: this color is hardcoded in many places 
	ctx.fillRect(0, 0, w, h);
}


// swap start and end points coordinates
function swap() {
  let tmp = startX;
	startX = endX;
  endX = tmp; 
  tmp= startY;
	startY= endY;
	endY= tmp;
}


/* html canvas => mousemove event	=> a pointing device is moved over an element (fired continuously as the mouse moves).
   this event is fired by html canvas element => document.getElementById('can') 
	 <canvas id="can" onmousemove="mouseMove(event)"></canvas>  */
function mouseMove(e)
{    	
	if (is_path) 
	{		
		let ctx= document.getElementById('can').getContext('2d'), x= e.clientX, y = e.clientY;		
		
		let px = ctx.getImageData(x, y, 1, 1).data; 
		
		// https://stackoverflow.com/questions/6735470/get-pixel-color-from-canvas-on-mousemove
		let not_ok=  ( px[0] == 229  &&  px[1] == 229  &&  px[2] == 229 ) ? true : false;  // #e5e5e5
		
		if ( not_ok )  {		  
			err++;
			is_path= false;
			gotoStart();   // return to start point
      drawCircle();			
		}		
		else  
		{				
			xMouse= x;   
			yMouse= y; 			
			if ( px[0] == 0  &&  px[1] == 0  &&  px[2] == 255 ) {  // is it end of task? => is user reached blue circle ?
				point= 1;
				theEnd(false);
			}
		}
	}	
}


// html <canvas id="can" onmousemove="mouseMove(event)" onmousedown="mouseDown()"></canvas> 
// user must click on start (red) circle to white path be visible                      
function mouseDown(e)  
{	 			
	// check distance from target circle in px (current position in pixels). e.offsetX and e.offsetY => mouse positions x, y	
	if ( !is_path ) { 
		is_path= ( getDistance(e.offsetX, e.offsetY, xMouse, yMouse) <= radius ) ? true: false;
		if ( is_path )  {
		  drawPath();
		}
	}
}


// Long live Saint Pythagoras 
// return distance from point P1(x1,y1) to point P2(x2,y2)
function getDistance(x1, y1, x2, y2)  {
  return Math.sqrt( Math.pow(x1-x2, 2) + Math.pow(y1-y2, 2) ); // total line distance (px) 
}	
 
 
// Changes application settings that can be changed directly by the user within the options form
function setVarPar()  
{   
	
	if ( document.getElementById('tpath').selectedIndex > -1 ) {
		tpath= parseInt(document.getElementById('tpath').value, 10); 
	}
	
	if ( document.getElementById('direc').selectedIndex > -1 ) {
		direc= parseInt(document.getElementById('direc').value, 10); 
	}
	
	// window.radius is set latter in setPar()
		
	if (document.getElementById('period').validity.valid) {
		period= Math.round(parseInt(document.getElementById('period').value, 10));  
  }	
	
  // timer_enabled => update latter in startTask()
	
	if (document.getElementById('timer').validity.valid) {
		timer= Math.round(parseInt(document.getElementById('timer').value, 10));  // task time (s)
  }	
		
  store= (document.getElementById('store_group').disabled === false && document.getElementById('store').value === '1') ? true : false;
  
}


// check is there is enought space between white stripes only for Rectangle path type
function checkDel()  {
	
	let rad, ok= true; // rad => update local copy of window.radius
	if ( document.getElementById('radius').selectedIndex > -1 ) {  // update window.radius. convert percents into pixels
		let r = parseInt ( document.getElementById('radius').value, 10 );
		rad = Math.floor ( r * Math.min( window.innerWidth, window.innerHeight ) / 100 );  // must be integer => it is index in path array => see gotoStart()
	}	
	else {
		rad = radius;
	}
	
	if ( direc == 1 || direc == 2 )  {  // Left-Right or Right-Left
		// let dx = ( ( bounds.right - bounds.left) / period  - 4 * rad ) / 2; 
		let dx = ( ( window.innerWidth - 2 * rad ) / period  - 4 * rad  ) / 2;
    if ( dx < 10 )	{
      ok = false;
		}			
	}
	else  {   // Up-Down or Down-Up
		//let dy = ( ( bounds.bottom - bounds.top) / period  - 4 * radius ) / 2; 
		let dy = ( (window.innerHeight - 2 * rad) / period  - 4 * rad ) / 2; 
		if ( dy < 10 )  {
			ok= false;
		}
	}

  if ( !ok )  {
    document.getElementById('is_valid').textContent='Circle size set too large and/or number of periods for your screen size. The path will not be clearly visible. Reduce at least one of these 2 parameters.';		
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
	
}


// end of task (research)
// if (is_error== true) => page (window) was resized during task (when is_task==true). Only show html error warning
function theEnd(is_error)  
{     
  timer_enabled= false;
	is_task= false;  
  	
	clear();
	document.getElementById('can').style.display= 'none'; 
  document.getElementById('end').style.display= 'block';   

  if (is_error) {
		document.getElementById('error').style.display= 'block';
		document.getElementById('result_wrapper').style.display= 'none';
		document.getElementById('score').textContent= '';
	}
	else 
	{
		document.getElementById('result_wrapper').style.display= 'block'; 
		document.getElementById('error').style.display= 'none';
		
		let endTime = new Date();   
		let t_time= Math.round(( endTime.getTime() - startTime.getTime() ) / 1000);  // seconds (convert time difference from milliseconds to seconds)	
		
		createTableArray(document.getElementById('table1'), [point, err, t_time],  ['Number of points', 'Number of errors', 'Task time (s)']);  		
		
		document.getElementById('table1').style.display= 'block';	
			
		// update 'score' label => above "Repeat" button
		document.getElementById('score').textContent= 'Number of points =' + point + '. Number of errors = ' + err + '. Task time (s) = ' + t_time; 
		
		document.getElementById('l_ts').textContent='Last completed research: ' + new Date();
		document.getElementById('l_res').style.display= 'block';
		document.getElementById('table_grp').style.display= 'none';
		try 
		{
			if ( store ) {
				setStorage(t_time);  
			}
		}        
		catch (e)  { 
			document.getElementById('table_grp').style.display= 'block';
			document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
		} 
	}		// if (!is_error)
  
}


// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	//if(dd<10) { dd='0'+dd; } 
	if ( mm < 10 ) { mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}

 
/* Store the results of the last completed task in the user's browser
   set user data into window.localStorage. Parameter:
	 time_s = Task time in seconds (from startTask() and theEnd() functions)  */  		
function setStorage(time_s)   
{    
	let oldItems = JSON.parse(localStorage.getItem('ptt-v1.0')) || []; 
  let id= oldItems.length+1;     	 
	// set new "record"
  let newItem = {
      'id' : id,
      'date' : setDate(),   // new Date(),  		         
			'path' : tpath,
      'direction' : direc,		
			'size': document.getElementById('radius').value,
		  'period' : (tpath == 1 || tpath == 2) ? period : '',		
			'max-time' : document.getElementById('timer_enabled').checked ? stringTime(timer) : '',
 			'time': time_s,
			'point' : point,
			'error' : err
  };

  oldItems.push(newItem);
  localStorage.setItem('ptt-v1.0', JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems);
  
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
            col.push(key);
        }
    }
  } 
  let table = document.createElement("table");
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement("th");      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= "";
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	let table= document.createElement("table");  // html table
  let tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (let i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement("th");      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (let i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (let j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (let i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined){
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


// clear web browser storage data for this application
function clrStorage()  
{  
  let cnt= localStorage.getItem('ptt-v1.0')  ?  JSON.parse(localStorage.getItem('ptt-v1.0')).length : 0;                                                        
  if (cnt === 0)  {
    alert('You have no data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved application results:  '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 { 
   window.localStorage.removeItem('ptt-v1.0'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data.';
	 document.getElementById('store_oper').style.display= 'none';
 }  
}


// copy data from user web browser to the Clipboard 
function copyStorage()  
{  
  let data= localStorage.getItem('ptt-v1.0');
	if (!data)  {
    alert('You have no data to Copy');
    return;
  }  
	data = JSON.parse(data); 
	let arr= jsonToArray(data);	
	copy2DToClipboard(arr);
}	


// convert json object into 2D array
// data = JSON object
// return 2D array
function jsonToArray(data)  
{
   // extract column (fields) names  
  let total= data.length, col = [], row, res=[];  // res= result (2d array) 
  for (let i = 0; i < total; i++) {
    for (let key in data[i]) {
			if (col.indexOf(key) === -1)  {
				col.push(key);  // col= columns (fields) names
			}
    }
  } 
	res.push(col);  // add field names as 1 row
  let c = col.length;    // column count  
  // add JSON data to the 2d array as rows.
  for (let i = 0; i < total; i++) 
  {
		row= [];
    for (let j = 0; j < c; j++)  {     
      row.push(data[i][col[j]]);
    }
		res.push(row);
  }
	return res;	
}


// https://stackoverflow.com/questions/60698115/how-to-copy-a-javascript-2d-array-to-clipboard-to-paste-it-in-excel
// copy 2D array to Clipboard
function copy2DToClipboard(arr) {
  let csv = '', row, cell;
  for (row = 0; row < arr.length; row++) {
    for (cell = 0; cell < arr[row].length; cell++) {
      csv += (arr[row][cell]+'').replace(/[\n\t]+/g, ' ');
      if (cell+1 < arr[row].length) csv += '\t';
    }
    if (row+1 < arr.length) { 
			csv += '\n'; 
		}
  }
  copyTextToClipboard(csv);
}


// copy 2D array (or text) to Clipboard
function copyTextToClipboard(txt) 
{	
	if (!navigator.clipboard) {
    fallbackCopy(txt);
    return;
  }
  navigator.clipboard.writeText(txt).then(function() {
  }, function(err) {
    alert('Oops, could not copy data. Error: ' + err);
  });
}


// source from: https://stackoverflow.com/questions/400212/how-do-i-copy-to-the-clipboard-in-javascript
// fallback copy text to clipboard
function fallbackCopy(txt) 
{
  let ta = document.createElement('textarea');
  ta.value = txt; 
  // Avoid scrolling to bottom
  ta.style.top = '0';
  ta.style.left = '0';
  ta.style.position = 'fixed';
  document.body.appendChild(ta);
  ta.focus();
  ta.select();
  try {
    let successful = document.execCommand('copy');
    let msg = successful ? 'successful' : 'unsuccessful';
  } catch (err) {
    alert('Oops, unable to copy. Error: ' + err);
  }
  document.body.removeChild(ta);
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays user saved results (Window.localStorage application data) in new empty web page (tab) 
function newWin()
{ 
	let data = JSON.parse(localStorage.getItem('ptt-v1.0'))
	createTableJSON(document.getElementById('table_grp'), data);
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= document.getElementById('table_grp').innerHTML;	
	let win = window.open('');
  win.document.write('<html><head><title>Saved cumulative data</title>' + stl + '</head><body>' + tbl + '</body></html>');
  win.focus();
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays Latest research results in new empty web page (tab) 
// parameter: id => DOM/html object (div or table)
function newWinLast(id)
{ 
  let e = document.getElementById(id);	
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= e.innerHTML; 	
	let win = window.open('');
	win.document.write('<html><head><title>Latest research data</title>' + stl + '</head><body>' + tbl + '</body></html>');
	win.focus();
}


// copy only latest task results (data) to the Clipboard 
function copyLast()  
{     
  let tbl = document.getElementById('table1').getElementsByTagName('table')[0];  // note: id='table1' is html div element
  if (tbl)  {
		let data= table2DArray(tbl);
		if (data.length > 0) {
			copy2DToClipboard(data);
		}
	}
}	


// https://stackoverflow.com/questions/3065342/how-do-i-iterate-through-table-rows-and-cells-in-javascript
// input argument tbl == html/DOM  table  object
// return 2D array
function table2DArray(tbl) 
{
	let res = [];	 
	for (let i = 0, row; row = tbl.rows[i]; i++) 
	{   //iterate through rows. rows would be accessed using the "row" variable assigned in the for loop
	  res[i]= [];		
		for (let j = 0, col; col = row.cells[j]; j++) {  //iterate through columns. columns would be accessed using the "col" variable assigned in the for loop			 
		  res[i][j]= col.textContent;
		}  
	}
	return res;				
}
